using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Input.Touch;
using GeometryHelper;

namespace SpinPaint
{
    public class Game1 : Microsoft.Xna.Framework.Game
    {
        GraphicsDeviceManager graphics;
        SpriteBatch spriteBatch;
        Texture2D texture;
        uint[] pixels;
        Vector2 displayCenter;
        Vector2 textureCenter;
        int radius;
        float currentAngle;
        float previousAngle;
        List<float> xCollection = new List<float>();

        class TouchInfo
        {
            public Vector2 PreviousPosition;
            public Vector2 CurrentPosition;
        }
        Dictionary<int, TouchInfo> touchDictionary = new Dictionary<int, TouchInfo>();

        public Game1()
        {
            graphics = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";
            graphics.IsFullScreen = true;       // TODO: Remove when ClientBounds is working!

            // Frame rate is 30 fps by default for Windows Phone.
            TargetElapsedTime = TimeSpan.FromTicks(333333);

            // Pre-autoscale settings.
            graphics.PreferredBackBufferWidth = 480;
            graphics.PreferredBackBufferHeight = 800;
        }

        protected override void Initialize()
        {
            base.Initialize();
        }

        protected override void LoadContent()
        {
            spriteBatch = new SpriteBatch(GraphicsDevice);

            Rectangle clientBounds = this.GraphicsDevice.Viewport.Bounds;
            int textureDimension = Math.Min(clientBounds.Width, clientBounds.Height);
            texture = new Texture2D(this.GraphicsDevice, textureDimension, textureDimension);
            pixels = new uint[texture.Width * texture.Height];
            radius = textureDimension / 2;
            textureCenter = new Vector2(radius, radius);
            displayCenter = new Vector2(clientBounds.Center.X, clientBounds.Center.Y);

            for (int y = 0; y < texture.Height; y++)
                for (int x = 0; x < texture.Width; x++)
                    if (IsWithinCircle(x, y))
                        pixels[y * texture.Width + x] = Color.White.PackedValue;

            texture.SetData<uint>(pixels);
        }

        protected override void Update(GameTime gameTime)
        {
            // Allows the game to exit
            if (GamePad.GetState(PlayerIndex.One).Buttons.Back == ButtonState.Pressed)
                this.Exit();

            // Disk rotates every 5 seconds
            double seconds = gameTime.TotalGameTime.TotalSeconds;
            currentAngle = (float)(2 * Math.PI * seconds / 5);

            // Colors cycle every 10 seconds
            float fraction = (float)(6 * (seconds % 10) / 10);
            Color clr;

            if (fraction < 1)
                clr = new Color(1, fraction, 0);
            else if (fraction < 2)
                clr = new Color(2 - fraction, 1, 0);
            else if (fraction < 3)
                clr = new Color(0, 1, fraction - 2);
            else if (fraction < 4)
                clr = new Color(0, 4 - fraction, 1);
            else if (fraction < 5)
                clr = new Color(fraction - 4, 0, 1);
            else
                clr = new Color(1, 0, 6 - fraction);

            // First assume no finger movement
            foreach (TouchInfo touchInfo in touchDictionary.Values)
                touchInfo.CurrentPosition = touchInfo.PreviousPosition;

            // Process touch
            TouchCollection toucheLocations = TouchPanel.GetState();

            foreach (TouchLocation touchLocation in toucheLocations)
            {
                int id = touchLocation.Id;

                switch (touchLocation.State)
                {
                    case TouchLocationState.Pressed:
                        if (!touchDictionary.ContainsKey(id))
                            touchDictionary.Add(id, new TouchInfo());

                        touchDictionary[id].PreviousPosition = TranslateToTexture(touchLocation.Position);
                        touchDictionary[id].CurrentPosition = TranslateToTexture(touchLocation.Position);
                        break;

                    case TouchLocationState.Moved:
                        if (touchDictionary.ContainsKey(id))
                            touchDictionary[id].CurrentPosition = TranslateToTexture(touchLocation.Position);
                        break;

                    case TouchLocationState.Released:
                        if (touchDictionary.ContainsKey(id))
                            touchDictionary.Remove(id);
                        break;
                }
            }

            // Transform previous and current positions
            Matrix translate1 = Matrix.CreateTranslation(-textureCenter.X, -textureCenter.Y, 0);
            Matrix translate2 = Matrix.CreateTranslation(textureCenter.X, textureCenter.Y, 0);

            Matrix previousRotation = translate1 * Matrix.CreateRotationZ(-previousAngle) * translate2;
            Matrix currentRotation = translate1 * Matrix.CreateRotationZ(-currentAngle) * translate2;

            // Now draw from previous to current points
            bool textureNeedsUpdate = false;

            foreach (TouchInfo touchInfo in touchDictionary.Values)
            {
                Vector2 point1 = Vector2.Transform(touchInfo.PreviousPosition, previousRotation);
                Vector2 point2 = Vector2.Transform(touchInfo.CurrentPosition, currentRotation);
                float radius = 12;      // Sure hope pressure comes back!

                RoundCappedLine line = new RoundCappedLine(point1, point2, radius);

                int yMin = (int)(Math.Min(point1.Y, point2.Y) - radius);
                int yMax = (int)(Math.Max(point1.Y, point2.Y) + radius);

                yMin = Math.Max(0, Math.Min(texture.Height, yMin));
                yMax = Math.Max(0, Math.Min(texture.Height, yMax));

                for (int y = yMin; y < yMax; y++)
                {
                    xCollection.Clear();
                    line.GetAllX(y, xCollection);

                    if (xCollection.Count == 2)
                    {
                        int xMin = (int)(Math.Min(xCollection[0], xCollection[1]) + 0.5f);
                        int xMax = (int)(Math.Max(xCollection[0], xCollection[1]) + 0.5f);

                        xMin = Math.Max(0, Math.Min(texture.Width, xMin));
                        xMax = Math.Max(0, Math.Min(texture.Width, xMax));

                        for (int x = xMin; x < xMax; x++)
                        {
                            if (IsWithinCircle(x, y))
                            {
                                int xFlip = texture.Width - x;
                                int yFlip = texture.Height - y;

                                pixels[y * texture.Width + x] = 
                                pixels[y * texture.Width + xFlip] = 
                                pixels[yFlip * texture.Width + x] = 
                                pixels[yFlip * texture.Width + xFlip] = clr.PackedValue;
                            }
                        }
                        textureNeedsUpdate = true;
                    }
                }
            }

            if (textureNeedsUpdate)
            {
                // Not sure why this is needed here but not in PhingerPaint
                this.GraphicsDevice.Textures[0] = null;

                texture.SetData<uint>(pixels);
            }

            // Prepare for next time through
            foreach (TouchInfo touchInfo in touchDictionary.Values)
                touchInfo.PreviousPosition = touchInfo.CurrentPosition;
            
            previousAngle = currentAngle;

            base.Update(gameTime);
        }

        bool IsWithinCircle(int x, int y)
        {
            x -= texture.Width / 2;
            y -= texture.Height / 2;
            
            return x * x + y * y < radius * radius;
        }

        Vector2 TranslateToTexture(Vector2 point)
        {
            return point - displayCenter + textureCenter;
        }

        protected override void Draw(GameTime gameTime)
        {
            GraphicsDevice.Clear(Color.Navy);

            spriteBatch.Begin();
            spriteBatch.Draw(texture, displayCenter, null, Color.White, currentAngle, textureCenter, 1, SpriteEffects.None, 0);
            spriteBatch.End();

            base.Draw(gameTime);
        }
    }

}
